// <header>
/**
 *  Themamap: The Thematic Maps Production Tool
 *
 *  Copyright (c) 2008 Université de Caen Basse-Normandie, Certic
 *  Copyright (c) 2008 Université de Caen Basse-Normandie, GREYC
 *                     		
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  <http://www.gnu.org/licenses/>
 *
 *  See http://themamap.info.unicaen.fr/authors/
 *      for a list of contributors 
 */
// </header>
package fr.unicaen.openmap.app;

import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.IOException;
import java.net.MalformedURLException;

import javax.swing.JFrame;
import javax.swing.JOptionPane;
import javax.swing.UIManager;


import com.bbn.openmap.Environment;
import com.bbn.openmap.MapHandler;

import com.bbn.openmap.gui.MapPanel;
import com.bbn.openmap.gui.OpenMapFrame;
import com.bbn.openmap.util.ArgParser;
import com.bbn.openmap.util.Debug;

import fr.unicaen.openmap.PropertyHandler;
import fr.unicaen.openmap.gui.SplashScreen;
import fr.unicaen.openmap.gui.ThemaMapPanel;
import fr.unicaen.openmap.util.ImageExporter;

/**
 * The OpenMap application framework. This class creates a PropertyHandler that
 * searches the classpath, config directory and user's home directory for an
 * openmap.properties file, and creates the application based on the contents of
 * the properties files. It also creates an MapPanel and an OpenMapFrame to be
 * used for the application and adds them to the MapHandler contained in the
 * MapPanel. All other components are added to that MapHandler as well, and they
 * use the MapHandler to locate, connect and communicate with each other.
 */
public class ThemaMap {

    protected static MapPanel mapPanel;

    /**
     * Create a new OpenMap framework object - creates a MapPanel, OpenMapFrame,
     * and brings up the layer palettes that are being told to be open at
     * startup. The MapPanel will create a PropertiesHandler that will search
     * for an openmap.properties file.
     */
    public ThemaMap() {
        this(null);
    }

    /**
     * Create a new OpenMap framework object - creates a MapPanel, OpenMapFrame,
     * and brings up the layer palettes that are being told to be open at
     * startup. The properties in the PropertyHandler will be used to configure
     * the application. PropertyHandler may be null.
     */
    public ThemaMap(PropertyHandler propertyHandler) {
        Debug.output(new fr.unicaen.openmap.util.AboutMenuItemBuildVersion().getCopyrightMessage());
        ////mapPanel = new ThemaMapPanel(propertyHandler, true);
        mapPanel = new ThemaMapPanel(propertyHandler, true);
        //System.out.println("propertyHandler " +propertyHandler );
        // Schedule a job for the event-dispatching thread:
        // creating and showing this application's GUI.
        javax.swing.SwingUtilities.invokeLater(new Runnable() {

            public void run() {
                // TODO There's something going on here with the progress
                // reporter and the swing thread that is causing the app to hang
                // in Leopard.
                showInFrame();
            }
        });
    }

    protected void showInFrame() {
////    	 etc.
//    	try {
//    	   UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
//    	   //SwingUtilities.updateComponentTreeUI(frame);
//    	   //force chaque composant de la fenêtre à appeler sa méthode updateUI
//    	} catch (InstantiationException e) {
//    	} catch (ClassNotFoundException e) {
//    	} catch (UnsupportedLookAndFeelException e) {
//    	} catch (IllegalAccessException e) {}

        ((ThemaMapPanel) mapPanel).create();
        OpenMapFrame omf = new OpenMapFrame(true);
        //omf.setExtendedState(JFrame.MAXIMIZED_BOTH);
        setWindowListenerOnFrame(omf);
        getMapHandler().add(omf);
        ImageExporter.initMapHandler(getMapHandler());
        //SVGExporter.initMapHandler(getMapHandler());//must be done here?
        omf.setVisible(true);
        mapPanel.getMapBean().showLayerPalettes();
        Debug.message("basic", "OpenMap: READY");
        omf.setSize(omf.getToolkit().getScreenSize());
    }

    /**
     * A method called internally to set the WindowListener behavior on an OpenMapFrame
     * used for the OpenMap application. By default, this method adds a
     * WindowAdapter that calls System.exit(0), killing java. You can extend
     * this to add a WindowListener to the OpenMapFrame that does nothing or
     * something else.
     */
    protected void setWindowListenerOnFrame(OpenMapFrame omf) {
        omf.setDefaultCloseOperation(JFrame.DO_NOTHING_ON_CLOSE);
        omf.addWindowListener(new WindowAdapter() {

            public void windowClosing(WindowEvent e) {
                if (JOptionPane.showConfirmDialog(
                        ((ThemaMapPanel) mapPanel), Environment.getI18n().get(ThemaMapPanel.class, "close_confirm", "Quit ThemaMap?")) == JOptionPane.YES_OPTION) {
                    System.exit(0);
                }
            }
        });
    }

    /**
     * Get the MapHandler used for the OpenMap object.
     */
    public static MapHandler getMapHandler() {
        return mapPanel.getMapHandler();
    }

    /**
     * Get the MapPanel, the container for the OpenMap components.
     */
    public static MapPanel getMapPanel() {
        return mapPanel;
    }

    /**
     * Create and return an OpenMap object that uses a standard PropertyHandler
     * to configure itself. The OpenMap object has a MapHandler that you can use
     * to gain access to all the components.
     * 
     * @return OpenMap
     * @see #getMapHandler
     */
    public static ThemaMap create() {
        return new ThemaMap(null);
    }

    /**
     * Create and return an OpenMap object that uses a standard PropertyHandler
     * to configure itself. The OpenMap object has a MapHandler that you can use
     * to gain access to all the components.
     * 
     * @return OpenMap
     * @see #getMapHandler
     */
    public static ThemaMap create(String propertiesFile) {
        Debug.init();
        //System.out.println(propertiesFile);
        PropertyHandler propertyHandler = null;

        if (propertiesFile != null) {
            try {
                propertyHandler = new PropertyHandler(propertiesFile);
            } catch (MalformedURLException murle) {
                Debug.error(murle.getMessage());
                murle.printStackTrace();
                propertyHandler = null;
            } catch (IOException ioe) {
                Debug.error(ioe.getMessage());
                ioe.printStackTrace();
                propertyHandler = null;
            }
        }

        return new ThemaMap(propertyHandler);
    }


        /**
         * The main OpenMap application.
         */
        /*public static void main(String args[]){
        ThemaMap th = ThemaMap.create("exemple.properties");
        }*/

    public static void main(String args[]) {
        try {
            //SplashScreen splash = new SplashScreen();
            //splash.show();
            //UIManager.setLookAndFeel(new SubstanceCremeLookAndFeel());
            UIManager.put("swing.boldMetal", Boolean.FALSE);
            ArgParser ap = new ArgParser("OpenMap");
            String propArgs = null;
            ap.add("properties",
                    "A resource, file path or URL to properties file\n Ex: http://myhost.com/xyz.props or file:/myhome/abc.pro\n See Java Documentation for java.net.URL class for more details",
                    1);
            ap.parse(args);
            String[] arg = ap.getArgValues("properties");
            if (arg != null) {
                propArgs = arg[0];
            }
            ThemaMap.create(propArgs);
            //splash.close();
        } catch (Exception e) {
            e.printStackTrace();
        }


    }
//    public static exportMap(){
//
//        SavePropertiesFolderAsText.saveAs(null, layers, null, null);;
//    }
}
