// <header>
/**
 *  Themamap: The Thematic Maps Production Tool
 *
 *  Copyright (c) 2008 Université de Caen Basse-Normandie, Certic
 *  Copyright (c) 2008 Université de Caen Basse-Normandie, GREYC
 *                     		
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  <http://www.gnu.org/licenses/>
 *
 *  See http://themamap.info.unicaen.fr/authors/
 *      for a list of contributors 
 */
// </header>
package fr.unicaen.openmap.dataStorage;

import com.bbn.openmap.util.Debug;
// import com.bbn.openmap.util.CSVTokenizer;
import com.bbn.openmap.util.FileUtils;
import com.bbn.openmap.I18n;

import com.bbn.openmap.layer.location.LocationLayer;
import com.bbn.openmap.util.PropUtils;
import java.util.*;
import java.io.*;
import javax.swing.*;
import java.awt.event.*;

import fr.unicaen.openmap.util.Util;
import fr.unicaen.openmap.util.LonLatNames;
import fr.unicaen.openmap.util.propertyEditor.DataStoragePropertyEditor;
import fr.unicaen.openmap.util.propertyEditor.FileTypeDescription;
import fr.unicaen.openmap.dataStorage.gui.DataTableModel;
import fr.unicaen.openmap.dataStorage.gui.DataTableStoragePanel;
import fr.unicaen.openmap.util.propertyEditor.Inspector;
import java.awt.BorderLayout;
import java.awt.Color;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.util.zip.GZIPInputStream;
import java.net.URL;

/**
 * CSVFileDataStorage is adapted from com.bbn.openmap.io.CSVFile and uses 
 * an enhanced tokenizer derived from com.bbn.openmap.io.CSVTokenizer. 
 * <p>Field separators are comma, semicolon, vertical bar and tab. Comma is not a separator in case it is used
 * by the locale as decimal number separator.
 * <p>It always tries to read numbers as double, even when they use comma as decimal separator.
 * <p>Properties:
 * <ul>
 * <li><code>columnNamesExistInHeader</code> is also true by default and indicates that the first line column value are the column names;</li>
 * </ul>
 * idea: construct the localisation attribute by catenate the attribute starting with longitude and latitude.
 * <p>Example of properties:
 * <pre>
 * csvdstest.dataStorageClass=fr.unicaen.openmap.dataStorage.CSVFileDataStorage
 * csvdstest.csvFile=cities.csv
 * csvdstest.columnNamesExistInHeader=true
 * </pre>
 * Bug: don't take AttributesToLoad into account.
 * @author Jacques.Madelaine@info.unicaen.fr
 */
public class CSVFileDataStorage extends InformationDataStorage {

    public static final String CSV_FILE_BY_INFO = "csv.info";
    public static final String commaIsDelimiterPropertyString = "commaIsDelim";
    public static final String semicolonIsDelimiterPropertyString = "semicolonIsDelim";
    public static final String verticalbarIsDelimiterPropertyString = "verticalbarIsDelim";
    public static final String tabulationIsDelimiterPropertyString = "tabulationIsDelim";
    private ArrayList<DataObject> dataObjects;
    private ArrayList<String> columnNames;
    private ArrayList<String> typeNames;      // string, lon or lat
    private LonLatNames lonLatNames;    // suffixes of column names starting with longitude or latitude

    public CSVFileDataStorage(DataStorageListener dsl) {
        super(dsl);
        Debug.message("DataStorage", this + " CSVFileDataStorage(DataStorageListener dsl)");
        setCommaIsDelimiter(!decimalSeparatorIsComma());
    }
    protected boolean initialized = false;

    public String getName() {
        return "".equals(csvFile) ? name : stripFileNameWithClassPathDirectories(csvFile);
    }

    public boolean isReady() {
        return initialized;
    }
    private static final String csvFilePropertyString = "csvFile";
    private String csvFile = "";
    private static final String columnNamesExistInHeaderPropertyString = "namesInRow1";
    private boolean columnNamesExistInHeader = true;

    protected boolean getColumnNamesExistInHeader() {
        return columnNamesExistInHeader;
    }

    protected void setColumnNamesExistInHeader(boolean b) {
        columnNamesExistInHeader = b;
    }
    private boolean commaIsDelimiter = true;

    protected boolean getCommaIsDelimiter() {
        return commaIsDelimiter;
    }

    protected void setCommaIsDelimiter(boolean b) {
        commaIsDelimiter = b;
        //    System.err.println("CSVFile: "+csvFile + " " + " comma is " +  (commaIsDelimiter ? "" : "NOT") + " delimiter");
    }
    private boolean semicolonIsDelimiter = true;

    protected boolean getSemicolonIsDelimiter() {
        return semicolonIsDelimiter;
    }

    protected void setSemicolonIsDelimiter(boolean b) {
        semicolonIsDelimiter = b;
    }
    private boolean verticalbarIsDelimiter = true;

    protected boolean getVerticalbarIsDelimiter() {
        return verticalbarIsDelimiter;
    }

    protected void setVerticalbarIsDelimiter(boolean b) {
        verticalbarIsDelimiter = b;
    }
    private boolean tabulationIsDelimiter = true;

    protected boolean getTabulationIsDelimiter() {
        return tabulationIsDelimiter;
    }

    protected void setTabulationIsDelimiter(boolean b) {
        tabulationIsDelimiter = b;
    }
    private static final String delimitersPropertyString = "delimiters";
    //    private String delimiters;

    public String getDelimiters() {
        return (getCommaIsDelimiter() ? "," : "") +
                (getSemicolonIsDelimiter() ? ";" : "") +
                (getVerticalbarIsDelimiter() ? "|" : "") +
                (getTabulationIsDelimiter() ? "\t" : "");
    }
    // protected void setDelimiters(String d) { delimiters = d; }

    public void setProperties(String prefix, java.util.Properties properties) {
        // Debug.message("DataStorage" , "[CSVFileDataStorage] setProperties " + properties);
        prefix = PropUtils.getScopedPropertyPrefix(prefix);
        //        name = prefix;
        super.setProperties(prefix, properties);

        csvFile =
                properties.getProperty(prefix + csvFilePropertyString,
                //by PropertyInfo
                properties.getProperty(prefix + CSV_FILE_BY_INFO));
        columnNamesExistInHeader =
                Util.isTrue(properties.getProperty(prefix + columnNamesExistInHeaderPropertyString, "true"));
        commaIsDelimiter =
                Util.isTrue(properties.getProperty(prefix + commaIsDelimiterPropertyString, !decimalSeparatorIsComma() ? "true" : "false"));
        semicolonIsDelimiter =
                Util.isTrue(properties.getProperty(prefix + semicolonIsDelimiterPropertyString, "true"));
        verticalbarIsDelimiter =
                Util.isTrue(properties.getProperty(prefix + verticalbarIsDelimiterPropertyString, "true"));
        tabulationIsDelimiter =
                Util.isTrue(properties.getProperty(prefix + tabulationIsDelimiterPropertyString, "true"));

        Debug.message("DataStorage", "[CSVFileDataStorage] using file: " + csvFile + " with delimiters: [" + getDelimiters() + "] " + (columnNamesExistInHeader ? ", with  column names in header" : "") + ".");

        initialize();
    }

    public Properties getProperties(String prefix, Properties properties) {
        properties = super.getProperties(prefix, properties);
        prefix = PropUtils.getScopedPropertyPrefix(prefix);

        properties.put(prefix + csvFilePropertyString, stripFileNameWithClassPathDirectories(csvFile));
        properties.put(prefix + columnNamesExistInHeaderPropertyString, String.valueOf(columnNamesExistInHeader));
        //Used with getPropertyInfo
        properties.put(prefix + CSV_FILE_BY_INFO, //csvFilePropertyString,
                new FileTypeDescription(FileTypeDescription.CSV_DESCRIPTION, csvFile, FileTypeDescription.CSV_PATTERN));
        properties.put(prefix + commaIsDelimiterPropertyString, String.valueOf(getCommaIsDelimiter()));
        properties.put(prefix + semicolonIsDelimiterPropertyString, String.valueOf(getSemicolonIsDelimiter()));
        properties.put(prefix + verticalbarIsDelimiterPropertyString, String.valueOf(getVerticalbarIsDelimiter()));
        properties.put(prefix + tabulationIsDelimiterPropertyString, String.valueOf(getTabulationIsDelimiter()));

        return properties;
    }

    public Properties getPropertyInfo(Properties list) {
        list = super.getPropertyInfo(list);
        PropUtils.setI18NPropertyInfo(i18n,
                list,
                CSVFileDataStorage.class,
                CSV_FILE_BY_INFO,
                csvFilePropertyString, //label
                csvFilePropertyString, //tooltip
                "fr.unicaen.openmap.util.propertyEditor.FileTypePropertyEditor");
        PropUtils.setI18NPropertyInfo(i18n,
                list,
                CSVFileDataStorage.class,
                columnNamesExistInHeaderPropertyString,
                "names in 1rst row", //label
                "Does the first line contains columns names?", //tooltip
                "fr.unicaen.openmap.util.propertyEditor.CheckBoxPropertyEditor");
        /* problem as its value contanins separators!!!!
        PropUtils.setI18NPropertyInfo(i18n,
        list,
        CSVFileDataStorage.class,
        DELIMITERS_BY_INFO,
        "delimiters", //label
        "delimiters", //tooltip
        "fr.unicaen.openmap.util.propertyEditor.DelimitersPropertyEditor"
        );
         */
        PropUtils.setI18NPropertyInfo(i18n,
                list,
                CSVFileDataStorage.class,
                commaIsDelimiterPropertyString,
                "Comma is separator", //label
                "Comma is a field separator", //tooltip
                "fr.unicaen.openmap.util.propertyEditor.CheckBoxPropertyEditor");

        PropUtils.setI18NPropertyInfo(i18n,
                list,
                CSVFileDataStorage.class,
                semicolonIsDelimiterPropertyString,
                "Semicolon is separator", //label
                "Semicolon is a field separator", //tooltip
                "fr.unicaen.openmap.util.propertyEditor.CheckBoxPropertyEditor");

        PropUtils.setI18NPropertyInfo(i18n,
                list,
                CSVFileDataStorage.class,
                verticalbarIsDelimiterPropertyString,
                "Vertical bar is separator", //label
                "Vertical bar is a field separator", //tooltip
                "fr.unicaen.openmap.util.propertyEditor.CheckBoxPropertyEditor");

        PropUtils.setI18NPropertyInfo(i18n,
                list,
                CSVFileDataStorage.class,
                tabulationIsDelimiterPropertyString,
                "Tabulation is separator", //label
                "Tabulation is a field separator", //tooltip
                "fr.unicaen.openmap.util.propertyEditor.CheckBoxPropertyEditor");

        list.put(initPropertiesProperty,
                //		 getInitPropertiesPropertyString()
                CSV_FILE_BY_INFO + " " + columnNamesExistInHeaderPropertyString + " " +
                commaIsDelimiterPropertyString + " " + semicolonIsDelimiterPropertyString + " " +
                verticalbarIsDelimiterPropertyString + " " + tabulationIsDelimiterPropertyString);

        return list;
    }

    public String getInitPropertiesPropertyString() {
        return csvFilePropertyString;
    }

    /** get ith column name. Automagically set it if lacking. 
     * Warning column names must be retrieved in the right order for the first time, 
     * as they are set in that order! 
     */
    private String getColumnName(int i) {
        if (i >= columnNames.size()) {
            columnNames.add("Attr " + i);
        }
        return columnNames.get(i);
    }

    private String getTypeName(int i) {
        if (i >= typeNames.size()) {
            typeNames.add("string");
        }
        return typeNames.get(i);
    }

    /** try to figure out, from Locale, if the decimal number separator is a comma */
    static boolean decimalSeparatorIsComma() {
        return java.text.NumberFormat.getInstance().format(1.1).indexOf(',') != -1;
    }

    public void initialize() {
        initialized = false;
        dataObjects = new ArrayList<DataObject>();
        columnNames = new ArrayList<String>();
        typeNames = new ArrayList<String>();
        lonLatNames = new LonLatNames();
        InputStream in = null;
        in = Util.open(csvFile, true);
        //LocationLayer l = new LocationLayer();

        Inspector inspector = new Inspector();
        if (in == null) {
            do {
                JPanel p = new JPanel(new BorderLayout());
                String f = i18n.get(this, "file", "The file") + " ";
                String nf = " " + i18n.get(this, "not_found", "was not found - please browse to select a new one");
                JLabel lab = new JLabel("<html><b>" + f + "'" + csvFile + "'" + nf + "</b></html>");
                lab.setForeground(Color.RED);
                p.add(lab, BorderLayout.NORTH);
                csvFile = "#";
                p.add(DataStoragePropertyEditor.provideSimpleCustomPanel(this, inspector), BorderLayout.CENTER);
                int choice = JOptionPane.showConfirmDialog(null, p, "Please Select a CSV File",
                        JOptionPane.OK_CANCEL_OPTION,
                        JOptionPane.INFORMATION_MESSAGE);
                if (choice == JOptionPane.CANCEL_OPTION) {
                    return;
                }
                setProperties(inspector.collectProperties());

                //initialized = true;
            } while (!isReady());
            return;
        }



        CSVTokenizer csvt = new CSVTokenizer(new BufferedReader(new InputStreamReader(in)),
                getDelimiters());
        Debug.message("DataStorage", this + " " + csvt);

        Object token = csvt.token();


        if (columnNamesExistInHeader) { // read header column names
            for (int i = 0; !csvt.isNewline(token); i++) {
                if (token == null) {
                    token = "";
                }
                String colName = token.toString().trim();
                columnNames.add(colName);
                String type = lonLatNames.getLonLatType(colName);
                if (type != null) // we have a lon or lat column name
                {
                    typeNames.add(type);
                } else {
                    typeNames.add("string");
                }

                token = csvt.token();
                if (csvt.isEOF(token)) {
                    System.err.println("CSVFileDataStorage." + name + ": warning eof found while in line. Last line skipped");
                    break;
                }
            }

            Debug.message("DataStorage", "CSVFileDataStorage: headers => " + columnNames + " " + lonLatNames);

            token = csvt.token(); // eat newline
        }

        for (int lineCount = columnNamesExistInHeader ? 1 : 0; !csvt.isEOF(token); token = csvt.token(), lineCount++) {

            DataObject dataObject = new DataObject();

            for (int i = 0; !csvt.isNewline(token); i++) {
                String type = getTypeName(i);

                if (token == null) {
                    token = "";
                }

                dataObject.addData(getColumnName(i).toString(), token);

                try {
                    if (LonLatNames.LON.equals(type)) {
                        lonLatNames.putLon(i, token);
                    } else if (LonLatNames.LAT.equals(type)) {
                        lonLatNames.putLat(i, token);
                    }
                } catch (java.lang.ClassCastException e) {
                    System.err.println("*** CSVFileDataStorage bad cast to Double line " + lineCount + " for " + type + " \"" + token + "\"");
                } catch (java.lang.NumberFormatException e) {
                    System.err.println("*** CSVFileDataStorage number format exception line " + lineCount + " for " + type + " \"" + token + "\"");
                }

                token = csvt.token();
                if (csvt.isEOF(token)) {
                    System.err.println("CSVFileDataStorage: warning eof found while in line. Last line skipped");
                    break;
                }
            }

            // gather the centroids
            lonLatNames.putCentroids(dataObject);

            // System.out.println("# "+dataObject);
            dataObjects.add(dataObject);
        }
        dataTable.setTable(dataObjects,
                getDataAttributesNamesToDisplay(columnNames),
                lonLatNames.getOMGraphicNames(),
                getDataAttributesNamesAliases());

        // data are stored, we can create dataTableModel and panel to view them.
        dataTableModel = new DataTableModel(dataTable);
        resetDataTableStoragePanel(dataTableModel);
        initialized = true;
        dsl.dataStorageEvent(this);

    }

//    private InputStream open(String fileName, boolean tryToUnzip) {
//
//        InputStream in = null;
//        URL url = null;
//        try {
//            url = PropUtils.getResourceOrFileOrURL(fileName);
//        } catch (java.net.MalformedURLException e) {
//        }
//
//        if (url != null) {
//            try {
//                in = url.openStream();
//            } catch (java.io.IOException e) {
//                System.err.println("CSVFileDataStorage: IOException for file " + fileName);
//                in = null;
//            }
//        }
//
//        if (in != null && tryToUnzip) { // try to uncompress...
//            try {
//                in = new GZIPInputStream(in);
//            } catch (Exception notGZIPexception) {
//                // It's not a zip file, then we re-open the url
//                try {
//                    in = url.openStream();
//                } catch (java.io.IOException e) {
//                    System.err.println("CSVFileDataStorage: IOException for file " + fileName);
//                    in = null;
//                }
//            }
//        }
//
//        Debug.message("DataStorage", this + " <== CSVFileDataStorage.open(" + fileName + ")");
//        return in;
//    }
    public JComponent getGUI() {
//
//        Box top = Box.createHorizontalBox();
//        //	top.add(columnNamesExistInHeaderCB);
//
//        Box bottom = Box.createHorizontalBox();
//        //	bottom.add(csvLabel);
//        //	bottom.add(csvButton);
//
        JPanel panel = new JPanel();
        panel.add(DataStoragePropertyEditor.provideCustomPanel(this));
//        panel.add(top);
//        panel.add(bottom);
//        //return new ClosablePanel(" CSV: "+csvFile, panel);
        return panel;
    }
}

