// <header>
/**
 *  Themamap: The Thematic Maps Production Tool
 *
 *  Copyright (c) 2008 Université de Caen Basse-Normandie, Certic
 *  Copyright (c) 2008 Université de Caen Basse-Normandie, GREYC
 *                     		
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  <http://www.gnu.org/licenses/>
 *
 *  See http://themamap.info.unicaen.fr/authors/
 *      for a list of contributors 
 */
// </header>
package fr.unicaen.openmap.dataStorage.postgis.db;

import fr.unicaen.openmap.dataStorage.*;

import com.bbn.openmap.omGraphics.OMGraphic;
import com.bbn.openmap.util.Debug;
import fr.unicaen.openmap.dataStorage.postgis.PostgisDataStorage;
import java.sql.*;
import java.util.*;

/**
 * The thread used to retrieve data. Create OMgraphic from PostGISGeom objects.
 */
public class DBQueryThread extends Thread {
    
    //----------------------------------------------------------------------
    // Private data
    //----------------------------------------------------------------------
    private String errorString = "NO ERROR";
    private DBListener listener = null;
    private DBConnection dbConnection = null;
    private ArrayList <DataObject> dataObjects = null;
    private ArrayList<String> omGraphicsNames = null;
    private ArrayList<String> columnNames = null;
    private String query = null;
    
    //----------------------------------------------------------------------
    // Data access methods
    //----------------------------------------------------------------------
    public ArrayList <DataObject> getDataObjects() { return dataObjects; }
    public ArrayList <String> getOMGraphicsNames() { return omGraphicsNames; }
    public ArrayList <String> getColumnNames() { return columnNames; }
    public String getError() { return errorString; }
    
    /**
     * Constructor.
     */
    public DBQueryThread (DBConnection dbConnection, DBListener listener, String query) {
	this(dbConnection, listener, query, null);
    }
    
    public DBQueryThread (DBConnection dbConnection, DBListener listener, String query, ArrayList <DataObject> dataObjects) {
	this.listener = listener;
	this.dbConnection = dbConnection;
	this.query = query;
	this.dataObjects = dataObjects;
    }
    
    /**
     * This launches the thread.<br />
     * Go and find yourself in source code how it is working (or not :)
     */
    public void run () {
	if (dataObjects == null)
	    dataObjects = new ArrayList<DataObject>();
	if (dbConnection.getConnection() == null) {
	    Debug.message("DBQueryThread", "[DBQueryThread] executeQuery() : connection not initialized.");
	    errorString = "No database connection.";
	    if (listener != null) listener.stateChanged(PostgisDataStorage.DBState.QUERY_ERROR);
	    return;
	}
	try {
	    Statement s = dbConnection.getConnection().createStatement();
	    ResultSet result = s.executeQuery(query);

	    ResultSetMetaData rsmd = result.getMetaData();
	    
	    // first set the MetaData for this table.
	    columnNames = new ArrayList<String>();
	    omGraphicsNames = new ArrayList<String>();
	    
            for (int i = 1 ; i <= rsmd.getColumnCount() ; i++){
                if (rsmd.getColumnLabel(i).startsWith("geometry_") ||
                        rsmd.getColumnLabel(i).startsWith("boundingbox_") ||
                        rsmd.getColumnLabel(i).startsWith("centroid_")) {
                    String geometryLabel = rsmd.getColumnLabel(i).substring(rsmd.getColumnLabel(i).indexOf('_')+1);
                    if (! omGraphicsNames.contains(geometryLabel))
                        omGraphicsNames.add(geometryLabel);
                } else if (rsmd.getColumnLabel(i).startsWith("the_geom")) {
                    // we should use column type here
                    // nothing to do with incomplete geometry for the moment
                    // if (! omGraphicsNames.contains(rsmd.getColumnLabel(i)))
                        // omGraphicsNames.add(rsmd.getColumnLabel(i));
                } else {
                    columnNames.add(rsmd.getColumnLabel(i));
                }
            }
	    
	    // table.setColumnNames(v);
	    // table.setMetaData(rsmd);
	    
	    // clean up the data in the table (if already retrived);
	    dataObjects.clear();

	    int rowCount=0;
	    // then all the data.
	    
	    /*
	      COMMENT :
	      All this code should be reviewed. We should get indexes for all
	      geom, centr, boundbox ... and then, parse all the records of
	    */
	    while( result.next() ) {
		DBTableRow line = new DBTableRow();
		// add all the fields in the table.
		
		Hashtable<String,Object> datas = new Hashtable<String,Object>();
		Hashtable<String,OMDataLocation> graphics = new Hashtable<String,OMDataLocation>();
		
		for (int j = 1 ; j <= rsmd.getColumnCount() ; j++) {
		    Object object = result.getString(j);
		    // Get the column name
		    String label = rsmd.getColumnLabel(j);
		    if (label.startsWith("geometry_")) {
			String geometryName = getGeometryLabel("geometry_", label);
			if (graphics.get(geometryName) == null)
			    graphics.put(geometryName, new OMDataLocation ());
			((OMDataLocation) graphics.get(geometryName)).setOMGraphic(Factory.createOMGraphic(object.toString()));
		    } else if (label.startsWith("centroid_")) {
			String geometryName = getGeometryLabel("centroid_", label);
			if (graphics.get(geometryName) == null)
			    graphics.put(geometryName, new OMDataLocation ());
			((OMDataLocation) graphics.get(geometryName)).setCentroid(Factory.createCentroid(object.toString()));
		    } else if (label.startsWith("bounding_")) {
			String geometryName = getGeometryLabel("bounding_", label);
			if (graphics.get(geometryName) == null)
			    graphics.put(geometryName, new OMDataLocation ());
			((OMDataLocation) graphics.get(geometryName)).setBoundingBox(Factory.createBoundingBox(object.toString()));
		    } else if (label.startsWith("the_geom")) {
                        // nothing to do with incomplete geometry for the moment.
                        // TODO: use geometry with computed BB and Centroid
                    } else {
                        datas.put(label, object);
                    }
		}
		
		dataObjects.add(new DataObject(datas, graphics));
		// table.add(line);
	    }
	    s.close();
	    if (listener != null) listener.stateChanged(PostgisDataStorage.DBState.QUERY_OK);
	} catch( Exception e ) {
	    if (Debug.debugging("DBQueryThread")){
		e.printStackTrace();
	    }
	    errorString = e.getMessage();
	    if (listener != null) listener.stateChanged(PostgisDataStorage.DBState.QUERY_ERROR);

	    // Debug.message("DBQueryThread","lines in DBTable : "+table.getRowCount());

	    return;
	}
    }
    
    private String getGeometryLabel (String prefix, String label) {
	String res = label.substring(prefix.length());
	return res != "" ? res : "noname";
    }
}
