// <header>
/**
 *  Themamap: The Thematic Maps Production Tool
 *
 *  Copyright (c) 2008 Université de Caen Basse-Normandie, Certic
 *  Copyright (c) 2008 Université de Caen Basse-Normandie, GREYC
 *                     		
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  <http://www.gnu.org/licenses/>
 *
 *  See http://themamap.info.unicaen.fr/authors/
 *      for a list of contributors 
 */
// </header>
package fr.unicaen.openmap.layer.isochroneLayer;

import com.bbn.openmap.proj.coords.LatLonPoint;
import com.bbn.openmap.event.MapMouseListener;
import fr.unicaen.openmap.layer.dataDisplayLayer.*;
import fr.unicaen.openmap.dataStorage.*;
import fr.unicaen.openmap.util.*;
import java.util.*;

import java.awt.*;
import java.util.Properties;
import com.bbn.openmap.omGraphics.*;
import com.bbn.openmap.proj.*;
import com.bbn.openmap.util.quadtree.QuadTree;
import com.bbn.openmap.proj.ProjMath;
import com.bbn.openmap.util.ColorFactory;
import com.bbn.openmap.util.DataOrganizer;
import com.bbn.openmap.util.Debug;
import com.bbn.openmap.util.PropUtils;
import java.awt.event.InputEvent;
import java.awt.event.MouseEvent;

/*
  we shall suppress all bonud in road and node
 */

/**
 * This class is for computation and display of isochrone areas
 *
 * Exemples of properties
 * <pre>
 * ## isochrone
 * isochrone.class              = fr.unicaen.openmap.layer.isochroneLayer.IsochroneLayer
 * isochrone.dataStorageClass   = fr.unicaen.openmap.dataStorage.xmltable.XmlTableDataStorage
 * isochrone.prettyName         = Isochrone
 * isochrone.lineColor          = FF66FF66
 * isochrone.fillColor          = FFFFFFFF
 * isochrone.selectedLineColor  = FF990099
 * isochrone.selectedFillColor  = C900FF55
 * isochrone.fill               = yes
 *
 * #isochrone.xmlFile            = routesbessin.xml.gz
 * isochrone.xmlFile            = routes14.xml.gz
 * isochrone.geometryColumnName = route
 * ## classColumnName is used to decide the average speed of the road
 * isochrone.classColumnName    = numero
 * ## nodesPrecisionRange is used to decide if two nodes points are to be traited as identical
 * isochrone.nodesPrecisionRange= 0.001f
 * #isochrone.xmlFile            = routes14-2005.xml.gz
 * #isochrone.geometryColumnName = route
 * #isochrone.classColumnName    = NUM_ROUTE
 *
 * #isochrone.latOrigin          = 49.311f
 * #isochrone.lonOrigin          = -.829f
 * # courvaudon
 * #isochrone.latOrigin          = 49.04f
 * #isochrone.lonOrigin          = -.585f
 * #st germain la blanche herbe
 * isochrone.latOrigin          = 49.20f
 * isochrone.lonOrigin          = -.413f
 * isochrone.maxBounds          = 8
 * ## nbBounds actually used out of the maxBounds possible
 * isochrone.nbBounds           = 4
 * ## bounds time in minutes
 * isochrone.bound0             = 5
 * isochrone.bound1             = 10
 * isochrone.bound2             = 20
 * isochrone.bound3             = 30
 * isochrone.bound4             = 60
 * ## we should define at least maxBounds colors
 * isochrone.boundColor0        = FF00FF00
 * isochrone.boundColor1        = FFFFFF00
 * isochrone.boundColor2        = FFFF0000
 * isochrone.boundColor3        = FFFF00FF
 * isochrone.boundColor4        = FF00FFFF
 * isochrone.boundColor5        = FF00FF00
 * isochrone.boundColor6        = FFFFFF00
 * isochrone.boundColor7        = FFFF0000
 * </pre>
 */
public class IsochroneLayer extends SimpleDataDisplayLayer {
    
    // Properties names
    public static final String geometryColumnNamePropertyString = "geometryColumnName";
    public static final String classColumnNamePropertyString = "classColumnName";
    
    // why do we use theses configurations ?
    public static final String lineColorPropertyString = "lineColor";
    public static final String fillColorPropertyString = "fillColor";
    public static final String selectedFillColorPropertyString = "selectedFillColor";
    public static final String selectedLineColorPropertyString = "selectedLineColor";
    public static final String fillPropertyString = "fill";
    
    // these ones are more important but are optionnal
    public static final String latOriginPropertyString = "latOrigin";
    public static final String lonOriginPropertyString = "lonOrigin";
    public static final String maxBoundsPropertyString = "maxBounds";
    public static final String nbBoundsPropertyString = "nbBounds";
    public static final String boundPropertyString = "bound";
    public static final String boundColorPropertyString = "boundColor";
    public static final String nodesPrecisionRangePropertyString = "nodesPrecisionRange";
    
    public static final String autorouteSpeedString = "autoroute";
    public static final String nationaleSpeedString = "nationale";
    public static final String departementaleSpeedString = "departementale";
    public static final String liaisonSpeedString = "liaison";
    
    
    private int autorouteSpeed = 0;
    private int nationaleSpeed = 0;
    private int departementaleSpeed = 0;
    private int liaisonSpeed = 0;
    
    private int autorouteDefaultSpeed = 0;
    private int nationaleDefaultSpeed = 0;
    private int departementaleDefaultSpeed = 0;
    private int liaisonDefaultSpeed = 0;
    
    // data storage column names
    private String geometryColumnName;
    private String classColumnName;
    
    protected Colors colors = new Colors();
    private float latOrigin;
    private float lonOrigin;
    
    private float nodesPrecisionRange = 0.001f; // used to decide if two nodes points are to be traited as identical
    
    private Vector<GraphStructContainer.Node>[] isochroneNodes = null;
    protected int maxBounds = 8;
    protected int nbBounds = 0;
    protected float[] bounds = null;
    
    protected Color[] boundsColors;
    
    private int nbSetDistance = 0; // in order to make stats
    
    public static java.text.DecimalFormat myDF = new java.text.DecimalFormat("#.##");
    /**
     * Datas Containers
     */
    protected GraphStructContainer graphStruct = null;
    // protected boolean firstRender = true;
    
    
    private OMGraphicList roadsGraphicList = null;
    
    public void setProperties(String prefix, Properties properties) {
        super.setProperties(prefix, properties);
        
        prefix = PropUtils.getScopedPropertyPrefix(prefix);
        
        geometryColumnName = properties.getProperty(prefix + geometryColumnNamePropertyString, "g");
        classColumnName = properties.getProperty(prefix + classColumnNamePropertyString, "c");
        nodesPrecisionRange = com.bbn.openmap.util.PropUtils.floatFromProperties(properties, prefix + nodesPrecisionRangePropertyString, 0.001f);
        // why use this color properties ???
        colors.setLineColor(properties.getProperty(prefix + lineColorPropertyString, "FF003333"));
        colors.setFillColor(properties.getProperty(prefix + fillColorPropertyString, "FF3366FF"));
        colors.setSelectedLineColor(properties.getProperty(prefix + selectedLineColorPropertyString, "FF990099"));
        colors.setSelectedFillColor(properties.getProperty(prefix + selectedFillColorPropertyString, "C900FF55"));
        colors.setFillMode(Util.isTrue(properties.getProperty(prefix + fillPropertyString, "true")));
        
        autorouteSpeed = com.bbn.openmap.util.PropUtils.intFromProperties(properties, prefix + autorouteSpeedString, 115);
        nationaleSpeed = com.bbn.openmap.util.PropUtils.intFromProperties(properties, prefix + nationaleSpeedString, 80);
        departementaleSpeed = com.bbn.openmap.util.PropUtils.intFromProperties(properties, prefix + departementaleSpeedString, 65);
        liaisonSpeed = com.bbn.openmap.util.PropUtils.intFromProperties(properties, prefix + liaisonSpeedString, 35);
        
        autorouteDefaultSpeed = autorouteSpeed;
        nationaleDefaultSpeed = nationaleSpeed;
        departementaleDefaultSpeed = departementaleSpeed;
        liaisonDefaultSpeed = liaisonSpeed;
        
        // make the default to the Bessin centroid (Normandy)
        latOrigin = com.bbn.openmap.util.PropUtils.floatFromProperties(properties, prefix + latOriginPropertyString, 49.3f);
        lonOrigin = com.bbn.openmap.util.PropUtils.floatFromProperties(properties, prefix + lonOriginPropertyString, -.85f);
        
        // bounds
        nbBounds = com.bbn.openmap.util.PropUtils.intFromProperties(properties, prefix + nbBoundsPropertyString, 1);
        maxBounds = com.bbn.openmap.util.PropUtils.intFromProperties(properties, prefix + nbBoundsPropertyString, nbBounds);
        if (maxBounds < nbBounds) {
            maxBounds = nbBounds;
            System.err.println("IsochroneLayer] ### Warning maxBounds < nbBounds. We use " + maxBounds);
        }
        bounds = new float[maxBounds];
        for (int i = 0 ; i < nbBounds; i++)
            bounds[i] = com.bbn.openmap.util.PropUtils.floatFromProperties(properties, prefix + boundPropertyString + i, 5.0f*i);
        
        //
        boundsColors = new Color[maxBounds];
        for (int i = 0 ; i < maxBounds; i++)
            boundsColors[i] = ColorFactory.parseColorFromProperties(properties, prefix + boundColorPropertyString + i, "FF55FF55", true);
    }
    
    /**
     *
     */
    public void setBounds(float bounds[]) {
        this.bounds = bounds;
        this.nbBounds = bounds.length;
    }
    
    /**
     *
     */
    public void setOrigin(float latOrigin, float lonOrigin) {
        this.latOrigin = latOrigin;
        this.lonOrigin = lonOrigin;
    }
    
    // useless to implement as it works fine without it !-)
    public Properties getProperties(Properties properties) {
        properties = super.getProperties(properties);
        
        String prefix = PropUtils.getScopedPropertyPrefix(this);
        
        properties.put(prefix + autorouteSpeedString, autorouteSpeed);
        properties.put(prefix + nationaleSpeedString, nationaleSpeed);
        properties.put(prefix + departementaleSpeedString, departementaleSpeed);
        properties.put(prefix + liaisonSpeedString, liaisonSpeed);
        
        
        properties.put(prefix + geometryColumnNamePropertyString, geometryColumnName);
        properties.put(prefix + classColumnNamePropertyString, classColumnName);
        
        properties.put(prefix + nodesPrecisionRangePropertyString, nodesPrecisionRange);
        properties.put(prefix + lineColorPropertyString, ColorFactory.getHexColorString(colors.getLineColor()));
        properties.put(prefix + fillColorPropertyString, ColorFactory.getHexColorString(colors.getFillColor()));
        properties.put(prefix + selectedLineColorPropertyString, ColorFactory.getHexColorString(colors.getSelectedLineColor()));
        properties.put(prefix + selectedFillColorPropertyString, ColorFactory.getHexColorString(colors.getSelectedFillColor()));
        properties.put(prefix + fillPropertyString, new Boolean(colors.getFillMode()).toString());
        
        properties.put(prefix + latOriginPropertyString, latOrigin);
        properties.put(prefix + lonOriginPropertyString, lonOrigin);
        properties.put(prefix + nbBoundsPropertyString + " ", nbBounds);
        properties.put(prefix + maxBoundsPropertyString, maxBounds);
        for (int i = 0 ; i < nbBounds; i++) properties.put(prefix + boundPropertyString + i + " ", bounds[i]);
        for (int i = 0 ; i < maxBounds; i++) properties.put(prefix + boundColorPropertyString + i, ColorFactory.getHexColorString(boundsColors[i]));
        return properties;
    }
    
    /**
     * we can construct an isochrone layer with some paramter given by another layer
     *
     * we can take road file and other informations but not the center or configuration properties like
     * speed or bounds ...
     *
     * this give us a minimum set of information to take the others in a configuration view and set them after ...
     *
     *
     **/
    public void setProperties2(String prefix, Properties properties){
        super.setProperties(prefix, properties);
        
        prefix = PropUtils.getScopedPropertyPrefix(prefix);
        
        geometryColumnName = properties.getProperty(prefix + geometryColumnNamePropertyString, "g");
        classColumnName = properties.getProperty(prefix + classColumnNamePropertyString, "c");
        nodesPrecisionRange = com.bbn.openmap.util.PropUtils.floatFromProperties(properties, prefix + nodesPrecisionRangePropertyString, 0.001f);
        autorouteSpeed = com.bbn.openmap.util.PropUtils.intFromProperties(properties, prefix + autorouteSpeedString, 115);
        nationaleSpeed = com.bbn.openmap.util.PropUtils.intFromProperties(properties, prefix + nationaleSpeedString, 80);
        departementaleSpeed = com.bbn.openmap.util.PropUtils.intFromProperties(properties, prefix + departementaleSpeedString, 65);
        liaisonSpeed = com.bbn.openmap.util.PropUtils.intFromProperties(properties, prefix + liaisonSpeedString, 35);
        
        autorouteDefaultSpeed = autorouteSpeed;
        nationaleDefaultSpeed = nationaleSpeed;
        departementaleDefaultSpeed = departementaleSpeed;
        liaisonDefaultSpeed = liaisonSpeed;        
        
        boundsColors = new Color[maxBounds];
        for (int i = 0 ; i < maxBounds; i++)
            boundsColors[i] = ColorFactory.parseColorFromProperties(properties, prefix + boundColorPropertyString + i, "FF55FF55", true);
        
    }
    // useless to implement as it works fine without it !-)
    public Properties getProperties2(Properties properties) {
        properties = super.getProperties(properties);
        
        String prefix = PropUtils.getScopedPropertyPrefix(this);
        
        properties.put(prefix + geometryColumnNamePropertyString, geometryColumnName);
        properties.put(prefix + classColumnNamePropertyString, classColumnName);
        properties.put(prefix + nodesPrecisionRangePropertyString, nodesPrecisionRange);
        properties.put(prefix + autorouteSpeedString, autorouteSpeed);
        properties.put(prefix + nationaleSpeedString, nationaleSpeed);
        properties.put(prefix + departementaleSpeedString, departementaleSpeed);
        properties.put(prefix + liaisonSpeedString, liaisonSpeed);
        for (int i = 0 ; i < maxBounds; i++) properties.put(prefix + boundColorPropertyString + i, ColorFactory.getHexColorString(boundsColors[i]));
        return properties;
    }
    
    // set the speed coefficient for computing
    public void setAutorouteSpeedCoef(int speed){
        autorouteSpeed=speed;
    }
    // set the speed coefficient for computing
    public void setNationaleSpeedCoef(int speed){
        nationaleSpeed=speed;
    }
    // set the speed coefficient for computing
    public void setDepartementaleSpeedCoef(int speed){
        departementaleSpeed=speed;
    }
    // set the speed coefficient for computing
    public void setLiaisonSpeedCoef(int speed){
        liaisonSpeed=speed;
    }
    
    // get the speed coefficient for computing
    public int getAutorouteSpeedCoef(){
        return autorouteSpeed;
    }
    // set the speed coefficient for computing
    public int getNationaleSpeedCoef(){
        return nationaleSpeed;
    }
    // set the speed coefficient for computing
    public int getDepartementaleSpeedCoef(){
        return departementaleSpeed;
    }
    // set the speed coefficient for computing
    public int getLiaisonSpeedCoef(){
        return liaisonSpeed;
    }
    
    // get the speed coefficient for computing
    public int getAutorouteDefaultSpeedCoef(){
        return autorouteDefaultSpeed;
    }
    // set the speed coefficient for computing
    public int getNationaleDefaultSpeedCoef(){
        return nationaleDefaultSpeed;
    }
    // set the speed coefficient for computing
    public int getDepartementaleDefaultSpeedCoef(){
        return departementaleDefaultSpeed;
    }
    // set the speed coefficient for computing
    public int getLiaisonDefaultSpeedCoef(){
        return liaisonDefaultSpeed;
    }
    
    /**
     * gives the average speed in km/minutes.
     * <p>This function should used properties data!
     */
    private float getSpeedCoefFromNumber(String number) {
        if (number == null || number.length()==0)
            return liaisonSpeed/60F;
        switch(number.charAt(0)) {
            case 'A':
                return autorouteSpeed/60F;
            case 'N':
                return nationaleSpeed/60F;
            case 'D':
                return departementaleSpeed/60F;
        }
        return liaisonSpeed/60F;
    }
    
    /**
     * returns the smallest i such that <code>distance &lt; bounds[i]</code>
     */
    private int getDistanceBound(float distance) {
        for (int i = 0; i < nbBounds; i++)
            if (distance < bounds[i])
                return i;
        return nbBounds;
    }

    public void animationModeChanged(MonoValuedFilterDataStorage ads) {
        throw new UnsupportedOperationException("Not supported yet.");
    }
    
    /**
     * The graph data structure representation of a road network
     *<p>TBD: Vectors nodes and roads should be ArrayList
     **/
    public class GraphStructContainer {
        private DataOrganizer   nodesQuadTree;    // container of nodes allowing geospatial selection
        private ArrayList<Node>    nodes;   // allow a sequential access to all nodes
        private ArrayList<Road>    roads;   // idem for roads
        private HashSet<Node>   nodesWithinHorizon; // list of all nodes within horizon (i.e. candidate to be marked with a bound limit)
        private int      	nbDuplicateNodesWithinHorizon; // mainly for stats
        
        GraphStructContainer() {
            nodesQuadTree = new QuadTree();
            nodes = new ArrayList<Node>();
            roads = new ArrayList<Road>();
            nodesWithinHorizon = new HashSet<Node>();
            nbDuplicateNodesWithinHorizon = 0;
            init();
        }
        
        public Node addNode(double lat, double lon) {
            Node node = new Node((float)lat,(float) lon);
            nodesQuadTree.put((float)lat,(float) lon, node);
            nodes.add(node);
            return node;
        }
        
        private Road addRoad(OMGraphicList omg, float time) {
            Road road = new Road(omg, time);
            roads.add(road);
            return road;
        }
        
        public Iterator<GraphStructContainer.Node> nodeIterator() {
            return nodes.iterator();
        }
        
        public Iterator<Road> roadIterator() {
            return roads.iterator();
        }
        
        /*
          public void reset() {
          nodesQuadTree.clear();
          nodes.clear();
          }
         */
        
        public Vector getNodes(float n, float w, float s, float e) {
            return nodesQuadTree.get(n,w,s,e);
        }
        
        public int getNodesSize() {
            return nodes.size();
        }
        
        public Node findClosest(double lat, double lon, double limit) {
            return (Node)nodesQuadTree.get((float)lat, (float)lon, limit);
        }
        
        public Node findClosest(double lat, double lon) {
            return (Node)nodesQuadTree.get((float)lat, (float)lon);
        }
        
        private void resetNodes() {
            Iterator<Node> it = nodes.iterator();
            while (it.hasNext())
                (it.next()).reset();
            nodesWithinHorizon.clear();
            nbDuplicateNodesWithinHorizon = 0;
        }
        
        /**
         * inits the structure from the DataStorageIterator
         */
        public void init() {
            dataStorageIterator.reset();
            
            int castexception = 0; // just to record bad casts
            
            int nbomg = 0, count = 0; // just for debug or test to reduce the amount to read
            for (count = 0; dataStorageIterator.hasNext() /*&& count < 4 */; count++) {
                
                DataObject data = dataStorageIterator.next();
                // System.out.print(" route "+ classColumnName + ": "+(String)data.getData(classColumnName));
                
                // we don't skip roads with empty classColumnName
                //if (((String)data.getData(classColumnName)).length()>0){
                OMDataLocation omdl = data.getOMDataLocation(geometryColumnName);
                OMGraphic omg = omdl.getOMGraphic();
                omg.setVisible(true);
                
                try { // in case of cast problem
                    // our data appears to be a list of OMGraphicList of 1 OMGraphicList
                    // we unwrap it until we not obtain an OMGraphicList
                    // we expect an OMLine or an OMPoly
                    OMGraphicList omgl = (OMGraphicList)omg;
                    nbomg++;
                    
                    while (omgl.size() == 1 && omgl.getOMGraphicAt(0) instanceof OMGraphicList){
                        omgl = (OMGraphicList)omgl.getOMGraphicAt(0);
                    }
                    
                    Node[] extremities = getNodes(omgl, true);
                    if (extremities == null) {
                        castexception++;
                        continue;
                    }
                    float time = IsochroneUtil.getLenthInKM(omgl) / getSpeedCoefFromNumber((String)data.getData(classColumnName));
                    Road road = addRoad(omgl, time);
                    
                    // we should duplicate the roads only if they are not one way roads. TBD
                    extremities[0].addSucc(road, extremities[1]);
                    extremities[1].addSucc(road, extremities[0]);
                } catch (ClassCastException e) {
                    castexception++;
                }
                //}
            }
            Debug.message("IsochroneLayer", "[IsochroneLayer]"
                    + " [ " + nbomg + " graphics on " + count + " => " +roads.size() + " roads, " + nodes.size() + " nodes ]"
                    + (castexception > 0 ? (" ###  encoutered " + castexception + " ClassCastException") : ""));
        }
        
        /**
         * Extract two extremities Nodes from the OMGraphicList
         * of OMLine or OMPoly.
         * <p>Actually reates Nodes if the argument <code>create</code> is true.
         */
        public Node[] getNodes(OMGraphicList omgl, boolean create) {
            
            double [] coords; // to hold lat/lon coordonates of extremities
            double lat1=0f, lon1=0f, lat2=0f,lon2=0f;
            if (omgl.getOMGraphicAt(0) instanceof OMLine){
                // Start point: first point of the first OMLine
                coords = ((OMLine)omgl.getOMGraphicAt(0)).getLL();
                lat1 = coords[0];
                lon1 = coords[1];
                
                // End point: last point of the last OMLine
                coords = ((OMLine)omgl.getOMGraphicAt(omgl.size() - 1)).getLL();
                lat2 = coords[2];
                lon2 = coords[3];
            } else if (omgl.getOMGraphicAt(0) instanceof OMPoly) {
                // Start point : first point of the first OMPoly
                OMPoly poly  = (OMPoly)omgl.getOMGraphicAt(0);
                coords=poly.getRawllpts();  // get all points (beware: in Radians !)
                lat1 = ProjMath.radToDeg(coords[0]);
                lon1 = ProjMath.radToDeg(coords[1]);
                
                // End point: last point of the last OMPoly
                poly = (OMPoly)omgl.getOMGraphicAt(omgl.size() - 1);
                coords=poly.getRawllpts();
                lat2 = ProjMath.radToDeg(coords[coords.length-2]);
                lon2 = ProjMath.radToDeg(coords[coords.length-1]);
            } else {
                Debug.message("IsochroneLayer", "[IsochroneLayer] ## skipping omgl");
                return null;
            }
            
            // create nodes for start and end point unless they already exist
            Node[] extremities = new Node[2];
            
            // closest node of start point
            extremities[0] = findClosest(lat1, lon1, nodesPrecisionRange);
            if (create && extremities[0] == null)
                extremities[0] = addNode(lat1,lon1);
            
            // closest node of end point
            extremities[1] = findClosest(lat2, lon2, nodesPrecisionRange);
            if (create && extremities[1] == null)
                extremities[1] = addNode(lat2,lon2);
            
            return extremities;
        }
        
        
        /**
         * Mark all nodes with the shortest distance (in time) from the given origin.
         * Stop at the horizon, i.e. distance &gt; largest bound.
         */
        private void markNodesWithDist(Node origin) {
            float d0 = origin.getDistance();
            // System.out.println("Origin: "+origin+" , nb successors : "+origin.nbSuccessors());
            for(int i = 0 ; i < origin.nbSuccessors(); i++) {
                RoadNodePair rn = origin.getNthSucc(i);
                float d1 = d0 + rn.road.getTime();
                Node n = rn.node;
                if(nodesWithinHorizon.add(n) == false)
                    nbDuplicateNodesWithinHorizon++;
                if (! n.setDistance(d1))
                    continue; // because we don't improve the distance
                if (d1 >= bounds[nbBounds-1])
                    continue; // we go out the horizon
                markNodesWithDist(n);
            }
        }
        
        /**
         *
         */
        public OMGraphicList generateIsochronesSets(float latOrigin, float lonOrigin, OMGraphicList omText) {
            Node origin = findClosest(latOrigin, lonOrigin);
            
            resetNodes(); // set all distance to + infinity
            origin.setDistance(0.0f);
            nbSetDistance = 0;
            
            markNodesWithDist(origin);
            
            // just iterate throw nodesWithinHorizon to construct the list of OMPoint inside each limit
            OMGraphicList sets[] = new OMGraphicList[nbBounds];
            for(int b = 0; b < nbBounds; b++)
                sets[b] = new OMGraphicList();
            
            for(Node node : nodesWithinHorizon) {
                int bound = getDistanceBound(node.getDistance());
                if (bound  < nbBounds) { // some nodes out of the horizon could have been inserted !
                    Debug.message("IsochroneLayer", "isochrone: " + bound + " " + node);
                    sets[bound].add(node.getOMPoint(boundsColors[bound]));
                    if (omText != null)
                        omText.add(node.getGraphic(boundsColors[bound]));
                }
            }
            
            StringBuffer counts = new StringBuffer();
            OMGraphicList omgraphics = new OMGraphicList();
            for (int i = 0; i < nbBounds; i++) {
                omgraphics.add(sets[i]);
                counts.append(" ["+i+"]="+sets[i].size());
            }
            Debug.message("IsochroneLayer", "[IsochroneLayer] generateIsochronesSets"
                    + " origin:" + origin + ", " + nbBounds + " bound(s), "
                    + nodesWithinHorizon.size() + " nodes within horizon ( "
                    + nbDuplicateNodesWithinHorizon + " duplicates tentatives), "
                    + nbSetDistance + " modif nodes / " + roads.size() + " roads + " + nodes.size() + " nodes"
                    + " => points sets " + counts);
            return omgraphics;
        }
        
        /**
         * Pair used as successor of Node
         */
        public class RoadNodePair {
            public Node node;
            public Road road;
            
            RoadNodePair(Road r, Node n) {
                road = r;
                node = n;
            }
        }
        
        /**
         * Road contains the OMGraphic representation
         * and data used to compute time travel
         */
        public class Road {
            private OMGraphicList  rep;
            private float      time;
            
            public Road(OMGraphicList road, float t) {
                rep = road;
                time = t;
                // show the time of each road for debug
                // OMLine oml = (OMLine)road.getOMGraphicAt(0);
                // float[] coords = oml.getLL();
                // road.add(new OMText(coords[2]+0.002f, coords[3]+0.002f, "["+time+"]", 1));
            }
            
            public OMGraphicList getGraphic() {
                return rep;
            }
            
            public float getTime() {
                return time;
            }
            
            public void setColor(Color c) {
                rep.setLinePaint(c);
            }
            
            public String toString() {
                Node[] extremities = getNodes(rep, false);
                return "{ "+extremities[0]+", "+extremities[1]+" }";
            }
        } // end class Road
        
        /**
         * A node in the roads network
         * contains a vector of road node pairs connected to it
         */
        public class Node {
            private ArrayList<RoadNodePair>  succ;     // nodes linked to the main node & roads
            private float          lat;      // latitude
            private float          lon;      // longitude
            private float          distance; // distance from the origine
            
            Node(){}
            
            Node(float lat, float lon) {
                this.lat = lat;
                this.lon = lon;
                succ = new ArrayList<RoadNodePair>();
                distance = Float.MAX_VALUE;
            }
            
            public String toString() {
                return "["+lat+", "+lon+"] "+ distance;
            }
            
            public float getLatitude() {
                return lat;
            }
            
            public float getLongitude() {
                return lon;
            }
            
            public int nbSuccessors() {
                return succ.size();
            }
            
            public void addSucc(Road road, Node node) {
                // <skip duplicate road> //  shall we allow two roads to the same node ? To Be Fixed
                // for (int i = 0; i < succ.size(); i++) {
                //     if (getNthSucc(i).node == node) {
                // 	Debug.message("IsochroneLayer", "Discards a road");
                // 	return;
                //     }
                // }
                // </skip duplicate road>
                succ.add(new RoadNodePair(road, node));
            }
            
            public RoadNodePair getNthSucc(int i) {
                return succ.get(i);
            }
            
            public float getDistance() {
                return distance;
            }
            
            public boolean setDistance(float d) {
                if (d < distance) {
                    nbSetDistance++; // for debugging and tracing purpose only. Can be safely discard.
                    distance = d;
                    return true;
                }
                return false;
            }
            
            public void reset() {
                distance = Float.MAX_VALUE;
            }
            
            public OMGraphic getGraphic(Color c) {
                String label = "[" + myDF.format(distance) +"]";
                OMText omt = new OMText(lat, lon, label, OMText.JUSTIFY_LEFT);
                if (c != null)
                    omt.setLinePaint(c);
                return omt;
            }
            
            public OMPoint getOMPoint(Color c) {
                OMPoint point = new OMPoint(lat, lon);
                point.setLinePaint(c);
                point.setFillPaint(c);
                return point;
            }
        } // end class Node
    } // end class GraphStructContainer
    
    synchronized public OMGraphicList generateGraphics() {
        Debug.message("IsochroneLayer", this + " generateGraphics()");
        
        OMGraphicList omgraphics = new OMGraphicList();
        
        // get projection
        // tbd: use the bounding box to select the graphics.
        Projection proj = getProjection();
        if (proj == null)
            return omgraphics;
        
        OMGraphicList nodesAsOMText = null;
        // nodesAsOMText = new OMGraphicList(); // to show nodes as OMText
        omgraphics.add(generateIsochronesSets(nodesAsOMText));
        
        if (Debug.debugging("IsochroneLayer")) {
            // add the origin
            omgraphics.add(graphStruct.findClosest(latOrigin, lonOrigin).getGraphic(java.awt.Color.magenta));
            
            if (nodesAsOMText != null) omgraphics.add(nodesAsOMText);
            
            // add roads
            omgraphics.add(roadsGraphicList);
            Debug.message("IsochroneLayer", this + " adding " + roadsGraphicList.size() + " roads.");
        }
        
        omgraphics.generate(proj);
        Debug.message("IsochroneLayer", this + " generate " + omgraphics.size() + " graphics.");
        return omgraphics;
    }
    
    public OMGraphicList generateIsochronesSets() {
        return generateIsochronesSets(null);
    }
    
    /**
     * generates graphics labels for nodes as text if nodesAsOMText id not null for debug purposes
     * <p> Returns an OMGraphicList made of <code>nbBounds</code> OMGraphicList representing an Isochrones Set.
     * <br> An Isochrone Set is the set of road nodes represented with OMPoint.
     */
    public OMGraphicList generateIsochronesSets(OMGraphicList nodesAsOMText) {
        // if (firstRender) {
        //     firstRender = false;
            Debug.message("IsochroneLayer", this + " first time graphic generation, could take some time ...");
            
            // No iterator == no data => return an empty list.
            if (dataStorageIterator == null) return new OMGraphicList();
            
            graphStruct = new GraphStructContainer();
            
            // init the road graphic list
            roadsGraphicList = new OMGraphicList();
            Iterator<GraphStructContainer.Road> it = graphStruct.roadIterator();
            while (it.hasNext()) {
                roadsGraphicList.add((it.next()).getGraphic());
            }
            Debug.message("IsochroneLayer", this + " graph structure created.");
        //}
        
        OMGraphicList isochroneSets = graphStruct.generateIsochronesSets(latOrigin, lonOrigin, nodesAsOMText);
        
        //	ArrayList<OMDataLocation> iso = generateIsochronesSetsAsOMDataLocation(isochroneSets);
        //	System.err.println(iso);
        
        Debug.message("IsochroneLayer", this + " " + isochroneSets.size() + " isochrone sets created.");
        return isochroneSets;
    }
    
    /**
     * will return null unless generateIsochronesSets has been called
     * (or generateIsochronesSetsAsOMDataLocation or generateGraphics that call generateIsochronesSets)
     */
    public OMGraphicList getRoadsGraphicList() {
        return roadsGraphicList;
    }
    
    /**
     * may be called instead of generateGraphics if the <code>IsochroneLayer</code>
     * is not used as a classical opemMap Layer.
     * <p> returns an arraylist of OMDataLocation, each of them representing an Isochrone set:
     * a bounding box and a OMGraphisList of OMPoint.
     * <p> If argument <code>isochroneSets</code> is null, they are computed with a call to
     * the method <code>generateIsochronesSets()</code>
     */
    public ArrayList<OMDataLocation> generateIsochronesSetsAsOMDataLocation(OMGraphicList isochroneSets) {
        if (isochroneSets == null)
            isochroneSets = generateIsochronesSets(null);
        ArrayList<OMDataLocation> isochroneOMDLSets = new ArrayList<OMDataLocation>();
        
        for (int i = 0; i < nbBounds; i++) {
            OMGraphicList omg = (OMGraphicList) isochroneSets.getOMGraphicAt(i);
            OMDataLocation omdl;
            if (omg.size() != 0) {
                OMPoint omp = (OMPoint) omg.getOMGraphicAt(0);
                BoundingBox bb = new BoundingBox(omp.getLon(), omp.getLat(), omp.getLon(), omp.getLat());
                for (Iterator it = omg.getTargets().iterator(); it.hasNext();) {
                    omp = (OMPoint) it.next();
                    bb.extendTo(omp.getLon(), omp.getLat());
                }
                omdl = new OMDataLocation(omg, null, bb);
            /*} else {
                omdl = new OMDataLocation(omg, null, new BoundingBox(0.0, 0.0, 0.0, 0.0));
            }*/
            isochroneOMDLSets.add(omdl);
            }
        }
        return isochroneOMDLSets;
    }
    
    public boolean isStorageReady(){
        if (dataStorageIterator==null ||dataStorageIterator.getDataStorage()==null||dataStorageIterator.getDataStorage().getCount()<=0) return false;
        return true;
    }
    
    //    private boolean hightScale(Projection proj) {
    //        return proj != null && proj.getScale() < 250000;
    //    }
    
    public static void printOMPoly(OMPoly omg, String mess) {
        double [] coords = omg.getRawllpts();
        StringBuffer sb = new StringBuffer();
        for (int i=0;i<coords.length;i++) {
            sb.append(ProjMath.radToDeg(coords[i])+",");
        }
        System.out.println(mess + "[ "+sb.toString()+" ]");
    }
    
    public static void printOMPoly(OMPoly omg) {
        printOMPoly(omg, "");
    }                  
}
