// <header>
/**
 *  Themamap: The Thematic Maps Production Tool
 *
 *  Copyright (c) 2008 Université de Caen Basse-Normandie, Certic
 *  Copyright (c) 2008 Université de Caen Basse-Normandie, GREYC
 *                     		
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  <http://www.gnu.org/licenses/>
 *
 *  See http://themamap.info.unicaen.fr/authors/
 *      for a list of contributors 
 */
// </header>
package fr.unicaen.openmap.util.classification;

import java.util.ArrayList;
import java.util.Collections;
import java.lang.reflect.Constructor;
import javax.swing.*;
import fr.unicaen.openmap.dataStorage.DataObject;
//import fr.unicaen.openmap.util.classification.gui.ClassificationPanel;
import fr.unicaen.openmap.util.Util;
/**
 * Perform a builtin statistical classification of data that must be doubles.
 * <p>You may get breaks and counts of objects per class.
 * <p>breaks[0] and breaks[breaks.length-1] are respectively equals to min and max values
 * <p>TODO: is it necessary to keep the <code>data</code> data member? If not it could be garbaged!
 */
public abstract class ClassificationOfDouble extends Classification {
    protected ArrayList<Double> data;
    
    public static final int NbClassesMax = 10;

    /**
     * @param data 
     * @param nbClasses the number of wished classes; some methods may decide to change its value.
     */
    public ClassificationOfDouble(ArrayList<Double> data, int nbClasses) {
	super(nbClasses);
	reInit(data); 
    }

    public ClassificationOfDouble(Classification c, int nbClasses) {
	super(nbClasses);
	reInit(c, nbClasses);
    }
    
    public void reInit(fr.unicaen.openmap.dataStorage.DataStorageIterator dsi,java.lang.String dataColumnName) {
	data = new ArrayList<Double>();
        dsi.reset();
	while (dsi.hasNext()) {
	    DataObject dataObject = dsi.next();
	    double value = Util.parseDouble(dataObject.getData(dataColumnName));
	    data.add(value);
	}
        reInit(data);
	dsi.reset();
	// System.out.println("### ClassificationOfString.reInit( "+ dataColumnName + " ) nbClasses = "+ nbClasses);

    }

    protected void reInit(ArrayList<Double> data, int nbClasses) {
	this.data = data;
	this.nbClasses = nbClasses;
	if (data != null)
	    init();
    }

    private void reInit(ArrayList<Double> data) {
	this.data = data;
	if (data != null)
	    init();
    }


    public void reInit(Classification c, int nbClasses) {
	ArrayList<Double> data;
	if (c instanceof ClassificationOfString) {
	    ClassificationOfString cos = (ClassificationOfString)c;
	    if (cos.data == null)
		data = null;
	    else {

		data = new ArrayList<Double>(cos.data.size());
		for (String s : cos.data)
		    data.add(Util.parseDouble(s));
	    }
	} else 	if (c instanceof ClassificationOfDouble) {
	    data = ((ClassificationOfDouble)c).data;
	} else {
	    System.err.println("### ClassificationOfDouble.reInit with an abstract Classification! ###");
	    return;
	}
	if (nbClasses > NbClassesMax)
	    nbClasses = NbClassesMax;
	reInit(data, nbClasses);
    }

    public boolean hasNoData() { return data == null || data.size() == 0; }

    public void unsetData() { data = null; } 

    protected double[] breaks = null;
    /**
     * breaks.size() == nbClasses + 1
     * <br>min == breaks[0] and max == breaks[nbClasses]
     */ 
    public double[] getBreaks() { return breaks; }
    public double getNthBreak(int n) { 
	return (breaks != null && n < breaks.length-1) ? breaks[n]  : 0; 
    }

//    protected int[] counts;
//    public int[] getCounts() { return counts; }
//    
    protected void initCountsFromBreaks() {
	counts = new int[breaks.length-1];		
	for(int i = 0; i < counts.length; i++)
	    counts[i] = 0;
		
	for(double d : data)
	    for(int j = 0; j < breaks.length-1; j++)
		if (d <= breaks[j+1]) {
		    counts[j]++;
		    break;
		}
    }


    /**
     * return the class rank of the value between 1 and nbClasses
     */
    public int getClass(Object o) {
	if (breaks == null)
	    return 0;
	double d = Util.parseDouble(o);
	int i = 1;
	for (; i <= nbClasses; i++)
	    if (d <= breaks[i])
		return i-1;
	return i-1;
    }

    public String toString() {

	String ret = "@"+hashCode()+" - ("+Util.className(this)+") ";
	int totalcounts = 0;
	ret += nbClasses + " [";
	//	System.out.println(ret);
	if (breaks != null && counts != null) {
	    for (int i = 0; i < breaks.length-1; i++) { 
		ret += String.format("%10.4f", breaks[i]) + "("+ counts[i] + "), ";
		totalcounts += counts[i];
	    }  
	    ret += breaks[breaks.length-1];
	}
	return ret + "] (" + totalcounts + ") " + (breaks != null ? breaks.length : "");
    }

    public String dump() {
	Collections.sort(data);
	String ret = "Data ("+data.size()+") [";
	for(Double d : data) ret += String.format("%10.4f; ", d);
	return ret + "]";
    }

    /** returns a text describing the ith class
     Format is not used!!
     */
    public String getClassText(int i, String format) {
	if (breaks != null)
	    if(i < breaks.length - 1)
		return ((i == 0 ? "[ " : "] ") +
			String.valueOf(breaks[i]) + " - " +
			String.valueOf(breaks[i+1]) +"]");
	
	return "";
    }

    /** returns texts describing the classes */
    public String[] getClassesTexts(String format) {
	int nbTexts = 0;
	if (breaks != null) //avoid null pointer
	    nbTexts = breaks.length - 1;
	String[] texts = new String[nbTexts];
	for (int i = 0; i < nbTexts; i++)
	    texts[i] = getClassText(i, format);
	return texts;
    }

    public String getBreaksAsString() {
	if (breaks == null) 
	    return "";
	String s = "";
	for (double d:breaks) {
	    s += "\t" + d;
	}
	return s;
    }
        
    /** standalone tests */
    public static void main(String[] args) {
	if (args.length < 3) {
	    System.err.println("usage: java fr.unicaen.openmap.util.classification.Classification <className> <nbClasses> <nbData> [<maxValue>]");
	    return;
	}

	ArrayList<Double> data = new ArrayList<Double>();
	String className = args[0];
	int n = Integer.parseInt(args[1]);
	int N = Integer.parseInt(args[2]);
	if (args.length > 4) {
	    N = args.length - 2;
	    for (int i = 2; i < args.length; i++)
		data.add(Double.parseDouble(args[i]));
	} else {
	    double maxValue = 70;
	    if (args.length == 4)
		maxValue = Double.parseDouble(args[3]);
	    
	    for (int i = 0; i < N; i++)
		data.add((Math.random()>0.25 ? +1 : -1)*(double)maxValue*Math.random());
	}
	ClassificationOfDouble c;
        try {
            Class cls = Class.forName("fr.unicaen.openmap.util.classification."+className);
	    Constructor[] constructors = cls.getConstructors();
	    // we should check for the signature of the constructors 
	    // see ThematicDataDisplayLayer.getConstructorWithStringParameter
	    c = (ClassificationOfDouble) constructors[0] .newInstance(new Object [] { data, n });;
        } catch (Exception e) {
            System.err.println("construction failure for "+ className);
	    e.printStackTrace();
	    return;
        }
        
	//System.out.println(c.dump());
	//System.out.println(c);
	int nb = 0, total = 0;
	int previouscc = 1;
	    //System.out.println("Value\tClasse\tCardinal");
	for (double d : data) {
	    int cc = c.getClass(d);
	    //System.out.format("%10.4f\t%d", d, cc);
	    if (cc == previouscc)
		nb++;
	    else {
		//System.out.print("\t" + nb);
		total += nb;
		previouscc = cc;
		nb = 1;
	    }
	    //System.out.println();
	}
	total += nb;
	//System.out.println("\t\t"+ nb);
	//System.out.println("Total\t\t"+ total);
	if (total != N)
	    	//System.out.println("### "+ total + " != " + N + "!!");
	return;
    }
}
